/*
 * l9wrapper.c - Level 9 interpreter DLL wrapper.
 * 
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include "l9wrapper.h"

/**
 * Array of .NET callback functions.
 */
#define OS_PRINTCHAR       0
#define OS_INPUT           1
#define OS_READCHAR        2
#define OS_FLUSH           3
#define OS_SAVE_FILE       4
#define OS_LOAD_FILE       5
#define OS_GET_GAME_FILE   6
#define OS_SET_FILENUMBER  7
#define OS_GRAPHICS        8
#define OS_CLEARGRAPHICS   9
#define OS_SETCOLOUR       10
#define OS_DRAWLINE        11
#define OS_FILL            12
#define OS_SHOW_BITMAP     13
FARPROC callback_ptr[14];

/**
 * The bitmap path for the current game. 
 */
char *bitmap_path = NULL;

/**
 * The bitmap type for the current game. 
 */
BitmapType bitmap_type = NO_BITMAPS;

/**
 * The currently displayed image.
 */
Bitmap *bitmap_data = NULL;

/**
 * The number of the currently displayed image.
 */
int bitmap_number = -1;

/**
 * Entry into dynamic-link library.
 * @param hInstance   handle to DLL module 
 * @param pvReserved  specifies further aspects of DLL initialization
 *                    and cleanup
 * @param fdwReason   specifies a flag indicating why the DLL entry-point
 *                    function is being called
 * @return TRUE if initializiation succeeds, otherwise FALSE
 */
int WINAPI DllMain (HINSTANCE hInstance, DWORD fdwReason, PVOID pvReserved)
{
	return TRUE;
}

/**
 * Sets the addresses of Level9.net frontend callback functions.
 * @param os_printchar       address of 'os_printchar' function 
 * @param os_input           address of 'os_input' function
 * @param os_readchar        address of 'os_readchar' function
 * @param os_flush           address of 'os_flush' function
 * @param os_save_file       address of 'os_save_file' function
 * @param os_load_file       address of 'os_load_file' function
 * @param os_get_game_file   address of 'os_get_game_file' function
 * @param os_set_filenumber  address of 'os_set_filenumber' function
 * @param os_graphics        address of 'os_graphics' function
 * @param os_cleargraphics   address of 'os_cleargraphics' function
 * @param os_setcolour       address of 'os_setcolour' function
 * @param os_drawline        address of 'os_drawline' function
 * @param os_fill            address of 'os_fill' function
 * @param os_show_bitmap     address of 'os_show_bitmap' function
 */
EXPORT void CALLBACK L9_SetCallbackFunctions(
	FARPROC os_printchar,
	FARPROC os_input,
	FARPROC os_readchar,
	FARPROC os_flush,
	FARPROC os_save_file,
	FARPROC os_load_file,
	FARPROC os_get_game_file,
	FARPROC os_set_filenumber,
	FARPROC os_graphics,
	FARPROC os_cleargraphics,
	FARPROC os_setcolour,
	FARPROC os_drawline,
	FARPROC os_fill,
	FARPROC os_show_bitmap)
{
	callback_ptr[OS_PRINTCHAR]      = os_printchar;
	callback_ptr[OS_INPUT]          = os_input;
	callback_ptr[OS_READCHAR]       = os_readchar;
	callback_ptr[OS_FLUSH]          = os_flush;
	callback_ptr[OS_SAVE_FILE]      = os_save_file;
	callback_ptr[OS_LOAD_FILE]      = os_load_file;
	callback_ptr[OS_GET_GAME_FILE]  = os_get_game_file;
	callback_ptr[OS_SET_FILENUMBER] = os_set_filenumber;
	callback_ptr[OS_GRAPHICS]       = os_graphics;
	callback_ptr[OS_CLEARGRAPHICS]  = os_cleargraphics;
	callback_ptr[OS_SETCOLOUR]      = os_setcolour;
	callback_ptr[OS_DRAWLINE]       = os_drawline;
	callback_ptr[OS_FILL]           = os_fill;
	callback_ptr[OS_SHOW_BITMAP]    = os_show_bitmap;
}

/**
 * os_printchar() prints a character to the output. The interface can either
 * buffer this character or print it immediately, but if buffering is used
 * then the characters must all be sent to the output when the interpreter
 * calls os_flush(). A paragraph of text is output as one long stream of
 * characters, without line breaks, so the interface must provide its own word
 * wrapping and any other features that are desired, such as justification or
 * a [More] prompt. The carriage return character is always '\r', rather than
 * '\n'.
 * @param c  the character to print
 */
void os_printchar(char c)
{
	callback_ptr[OS_PRINTCHAR](c);
}

/**
 * os_input() reads a line of text from the user, usually to accept the next
 * command to be sent to the game. The text input must be stored in ibuff with
 * a terminating zero, and be no longer than size characters. Normally
 * os_input() should return true, but may return false to cause the entire
 * input so far to be discarded. The reason for doing so is discussed in the
 * section at the end on allowing the interpreter to load a new game without
 * exiting.
 * @param ibuff  zero terminated input line
 * @param size   maximum length of input line
 * @returns FALSE causes the entire input to be discarded
 */
L9BOOL os_input(char *ibuff, int size)
{
	return callback_ptr[OS_INPUT](ibuff, size);
}

/**
 * os_readchar() looks to see if a key has been pressed if one has, returns the
 * character to the interpreter immediately. If no key has been pressed the
 * interpreter should wait for a key for at least the number of milliseconds
 * given in the argument. If after this period no key has been pressed, 0
 * should be returned. This is most commonly used when a game is exited,
 * causing it to print "Press SPACE to play again" and then call os_readchar().
 * @param millis  number of milliseconds to wait for a keystroke
 * @return 0 if no key has been pressed, otherwise the appropriate character
 */
char os_readchar(int millis)
{
	return callback_ptr[OS_READCHAR](millis);
}

/**
 * Called during dictionary listing. If true is returned (typically because
 * the user has pressed a key) then the listing is stopped. 
 * @return This routine should return immediately, without waiting. If this is
 * not possible then FALSE should be returned.
 */
L9BOOL os_stoplist(void)
{
	return 0;
}

/**
 * If the calls to os_printchar() are being buffered by the interface then the
 * buffered text must be printed when os_flush() is called.
 */
void os_flush(void)
{
	callback_ptr[OS_FLUSH]();
}

/**
 * os_save_file() should prompt the user in some way (with either text or a
 * file requester) for a filename to save the area of memory of size Bytes
 * pointed to by Ptr. true or false should be returned depending on whether or
 * not the operation was successful.
 * @param Ptr    area of memory to save
 * @param Bytes  number of bytes to save
 * @return TRUE if the operation was successful, otherwise false
 */
L9BOOL os_save_file(L9BYTE *Ptr, int Bytes)
{
	return callback_ptr[OS_SAVE_FILE](Ptr, Bytes);
}

/**
 * os_load_file() should prompt the user for the name of a file to load. At
 * most Max bytes should be loaded into the memory pointed to by Ptr, and the
 * number of bytes read should be placed into the variable pointed to by
 * Bytes.
 * @param Ptr    area of memory to write the data to
 * @param Bytes  number of bytes read
 * @param Max    maximum number of bytes to read
 * @return TRUE if the operation was successful, otherwise false
 */
L9BOOL os_load_file(L9BYTE *Ptr, int *Bytes, int Max)
{
	return callback_ptr[OS_LOAD_FILE](Ptr, Bytes, Max);
}

/**
 * os_get_game_file() should prompt the user for a new game file, to be stored
 * in NewName, which can take a maximum name of Size characters. When this
 * function is called the NewName array contains the name of the currently
 * loaded game, which can be used to derive a name to prompt the user with.
 * This is used by at least the Adrian Mole games, which load in the next part
 * of the game after the part currently being played has been completed.
 * These games were originally written for tape-based systems where the call
 * was simply "load the next game from the tape".
 * @param NewName  name of a game file
 * @param Size     maximum length of NewName
 * @return TRUE if the operation was successful. Otherwise false
 */
L9BOOL os_get_game_file(char *NewName, int Size)
{
	return callback_ptr[OS_GET_GAME_FILE](NewName, Size);
}

/**
 * os_set_filenumber() is for multi-part games originally written for
 * disk-based systems, which used game filenames such as gamedat1.dat,
 * gamedat2.dat etc. The routine should take the full filename in NewName
 * (of maximum size Size) and modify it to reflect the number n, e.g.
 * os_set_filename("gamedat1.dat",2) should leave "gamedat2.dat" in NewName.
 * @param NewName  name of game file
 * @param Size     maximum length of NewName
 * @param n        number to place in the game file's name
 */
void os_set_filenumber(char *NewName, int Size, int n)
{
	callback_ptr[OS_SET_FILENUMBER](NewName, Size, n);
}

/**
 * Called when graphics are turned on or off, either by the game or by the user
 * entering "graphics" or "text" as a command. If mode is 0 graphics should be
 * turned off, else they should be turned on. After a os_graphics(0) call all
 * the other graphics functions should do nothing.
 * Typically, if mode is not 0 the code will allocate some suitable bitmap for
 * drawing graphics into. To determine the size of the bitmap, the code should
 * call GetPictureSize(). The graphics routines should draw in a bitmap of the
 * size returned by this function, and then scale the bitmap appropriately for
 * display.
 * If instead the graphics code tries to scale the co-ordinates passed to
 * os_drawline() and os_fill() then problems occur with fill colours "leaking"
 * into other areas of the picture. The values returned by GetPictureSize()
 * will not change unless a new game is loaded.
 * The graphis bitmap is always 4 colour. The 4 colours are chosen from a
 * possible 8 by calls to os_setcolour (see below). Note that a call to
 * os_setcolour() must affect the colour of pixels already drawn on the bitmap.
 * For example, suppose a pixel in the bitmap is set to the first colour in the
 * palette during drawing, which at that moment is red. If later the first
 * colour in the palette is set to blue, at the end the pixel should be shown
 * blue. In order to actually draw graphics, the input routines os_input() and
 * os_readchar() should call RunGraphics(). This is discussed further below.
 * @param mode  0 if graphics are turned off, otherwise differs from 0
 */
void os_graphics(int mode)
{
	callback_ptr[OS_GRAPHICS](mode);
}

/**
 * Clear the current graphics bitmap by filling the entire bitmap with
 * colour 0.
 */
void os_cleargraphics(void)
{
	callback_ptr[OS_CLEARGRAPHICS]();
}

/**
 * Set the given colour in the graphics bitmap's palette to the colour at the
 * given index in the interpreter's table of colours. The actual table of
 * colours in the interpreters provided by Level 9 vary across different
 * machines. An acceptable palette that matches reasonably closely to the Amiga
 * releases is as follows (all colours are 8 bit R,G,B):
 * \code
 * 0x00,0x00,0x00  (black)
 * 0xFF,0x00,0x00  (red)
 * 0x30,0xE8,0x30  (green)
 * 0xFF,0xFF,0x00  (yellow)
 * 0x00,0x00,0xFF  (blue)
 * 0xA0,0x68,0x00  (brown)
 * 0x00,0xFF,0xFF  (cyan)
 * 0xFF,0xFF,0xFF  (white)
 * \endcode
 * @param colour  colour in the graphics bitmap's palette
 * @param index   index in the interpreter's table of colours
 */
void os_setcolour(int colour, int index)
{
	callback_ptr[OS_SETCOLOUR](colour, index);
}

/**
 * Draw a line on the graphics bitmap between (x1,y1) and (x2,y2). Note that
 * either point may lie outside of the bitmap, and that it is the
 * responsibility of the routine to clip to the appropriate coordinates.
 * For each point on the line, if the colour at that point is equal to colour2
 * the pixel's colour should be changed to colour1, else it should not be
 * modified.
 * @param x1       x coordinate of (x1,y1)
 * @param y1       y coordinate of (x1,y1)
 * @param x2       x coordinate of (x2,y2)
 * @param y2       y coordinate of (x2,y2)
 * @param colour1  drawing colour
 * @param colour2  colour to replace
 */
void os_drawline(int x1, int y1, int x2, int y2, int colour1, int colour2)
{
	callback_ptr[OS_DRAWLINE](x1, y1, x2, y2, colour1, colour2);
}

/**
 * If the pixel's colour at (x,y) is equal to colour2, fill the region
 * containing (x,y) with colour1. The boundaries of the region are defined as
 * those areas of the bitmap with a colour other than colour2.
 * @param x        x coordinate of pixel (x,y)
 * @param y        y coordinate of pixel (x,y)
 * @param colour1  fill colour
 * @param colour2  pixel colour
 */
void os_fill(int x, int y, int colour1, int colour2)
{
	callback_ptr[OS_FILL](x, y, colour1, colour2);
}

/**
 * Show the bitmap given by the number pic at the coordinates (x,y). Note that
 * the game can request the same picture several times in a row: it is a good
 * idea for ports to record the last picture number and check it against any
 * new requests.
 * @param pic  number of bitmap image
 * @param x    x coordinate on screen
 * @param y    y coordinate on screen
 */
void os_show_bitmap(int pic, int x, int y)
{
	if (pic != bitmap_number) {
		bitmap_data = DecodeBitmap(bitmap_path, bitmap_type, pic, x, y);
		bitmap_number = pic;
	}
	callback_ptr[OS_SHOW_BITMAP](pic, x, y);
}

/**
 * os_open_script_file() should prompt the user for the name of a
 * script file, from which input will be read until the end of the
 * script file is reached, and should return a pointer to the opened
 * script file, or NULL. This function is called in response to
 * the player entering the "#play" meta-command.
 * @returns pointer to an opened script file, or NULL
 */ 
FILE* os_open_script_file()
{
	return NULL;
}

/**
 * LoadGame() attempts to load a filename and then searches it for a valid
 * Level 9 game. If it is successful true is returned, else false. The previous
 * game in memory will be overwritten if the file filename can be loaded, even
 * if it does not contain a Level 9 game, so even if LoadGame() returns false
 * it must be assumed that the game memory has changed.
 * The second argument is the name of the file containing picture data, and may
 * be NULL. Ports should usually ask the user for just the filename and derive
 * picname from it in some way. The recommended approach is to first try the
 * filename with an extension of ".pic" and then try replacing the filename with
 * "picture.dat".
 * @param filename  file to load
 * @param picname   file containing picture data
 * @returns TRUE if successful, else FALSE
 */
EXPORT BOOL CALLBACK L9_LoadGame(char* filename, char* picname)
{
	BOOL success = LoadGame(filename, picname);

	if (success) {
		if (bitmap_path != NULL) {
			free(bitmap_path);
			bitmap_path = NULL;
		}
		bitmap_type = NO_BITMAPS;
		bitmap_number = -1;
	}

	return success;
}

/**
 * If LoadGame() has been successful, RunGame() can be called to run the
 * Level 9 game. Each call to RunGame() executes a single opcode of the game.
 * In pre-emptive multitasking systems or systems without any multitasking it
 * is enough to sit in a loop calling RunGame(), e.g.
 * \code
 * while (RunGame());
 * \endcode
 * @return TRUE if an opcode code was executed and FALSE if the game was
 *         stopped
 */
EXPORT BOOL CALLBACK L9_RunGame(void)
{
	return RunGame();
}

/**
 * StopGame() stops the current game from playing.
 */
EXPORT void CALLBACK L9_StopGame(void)
{
	StopGame();
}

/**
 * RestoreGame() attempts to restore the currently running game to the position
 * stored in the inFile saved game file. This gives interface code a means to
 * restore a game position.
 * @param inFile  saved game file
 */
EXPORT void CALLBACK L9_RestoreGame(char *inFile)
{
	RestoreGame(inFile);
}

/**
 * FreeMemory() frees memory used to store the game. This routine should be
 * called when exiting the interpreter.
 */
EXPORT void CALLBACK L9_FreeMemory(void)
{
	FreeMemory();
}

/**
 * Returns the width and height of the bitmap that graphics should be drawn
 * into. This is constant for any particular game.
 * @param width   width of bitmap
 * @param height  height of bitmap
 */
EXPORT void CALLBACK L9_GetPictureSize(int *width, int *height)
{
	GetPictureSize(width, height);
}

/**
 * Runs an opcode of the graphics routines. The simplest way to get graphics
 * to display is to add a loop to repeatedly call RunGraphics() to os_input()
 * and os_readchar():
 * \code
 *    while (RunGraphics());
 *    // Now draw graphics bitmap on display...
 * \endcode
 * Optionally, the code can provide a more "atmospheric" recreation of the
 * games by drawing the graphics slowly, as was the case on the old 8-bit
 * computers. This is achieved by calling RunGraphics() several times then
 * waiting for a while before calling it again. Note that when waiting the
 * code should still respond to user input.
 *
 * @return if a graphics opcode was run TRUE is returned, otherwise FALSE
 */
EXPORT BOOL CALLBACK L9_RunGraphics(void)
{
	return RunGraphics();
}

/**
 * Given a directory, returns the type of bitmap picture files in it, or
 * NO_BITMAPS if there are no bitmaps.
 * @param path  game directory
 * @return if supported pictures are found, TRUE is returned
 */
EXPORT BOOL CALLBACK L9_DetectBitmaps(char *path)
{
	if (path != NULL && (bitmap_path = strdup(path)) != NULL) {
		bitmap_type = DetectBitmaps(path);
	}
	return (bitmap_type == NO_BITMAPS ? FALSE : TRUE);
}

/**
 * This function loads and decodes the specified bitmap, returning a Bitmap
 * structure. The structure contains the width and height of the bitmap, a
 * palette of up to 32 colours used in the bitmap, and the actual data as an
 * array of indexes into the palette.
 * @param n  number of bitmap image
 * @return the loaded bitmap
 */
EXPORT Bitmap *CALLBACK L9_DecodeBitmap(int n)
{
	if (n == 0) {
		bitmap_data = DecodeBitmap(bitmap_path, bitmap_type, n, 0, 0);
		bitmap_number = 0;
	}
	return bitmap_data;
}
