/*
 * Config.cs - Manages the Level9.Net configuration.
 *
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Serialization;

namespace Level9
{
	/// <summary>
	/// Manages the Level9.Net configuration.
	/// </summary>
	public sealed class Config
	{
		// -- Constants ------------------------------------------------------

		/// <summary>
		/// The program name.
		/// </summary>
		public const string PROGRAM_NAME = "Level9.Net";

		/// <summary>
		/// The program version.
		/// </summary>
		public const string PROGRAM_VERSION = "0.9.6";

		/// <summary>
		/// The Level 9 interpreter DLL filename.
		/// </summary>
		public const string LEVEL9_DLL = "Level9.dll";

		/// <summary>
		/// The ANTLRv3 DLL filename.
		/// </summary>
		public const string ANTLR_DLL  = "Antlr3.Runtime.dll";

		/// <summary>
		/// The settings filename.
		/// </summary>
		public const string SETTINGS_FILE = "settings.xml";

		/// <summary>
		/// The log filename.
		/// </summary>
		public const string LOG_FILE = "level9net.log";

		/// <summary>
		/// The L9cut executable filename.
		/// </summary>
		public const string L9CUT_EXE = "L9cut.exe";

		/// <summary>
		/// The Level9.Net help filename.
		/// </summary>
		public const string HELP_FILE = "Level9Net.chm";

		/// <summary>
		/// The maximum number of 'recent files' in the main menu.
		/// </summary>
		const int MAX_RECENT_FILES = 10;

		// -- Structs, Enums, Inner Classes ----------------------------------

		/// <summary>
		/// The palette options for line drawn graphics.
		/// </summary>
		public enum Palette
		{
			/// <summary>User defined palette.</summary>
			UserDefined,
			/// <summary>Amiga palette.</summary>
			Amiga,
			/// <summary>Spectrum palette.</summary>
			Spectrum
		}

		/// <summary>
		/// Represents the size of a window.
		/// </summary>
		public class WindowSize
		{
			/// <summary>
			/// The window width.
			/// </summary>
			int width;

			/// <summary>
			/// The window height.
			/// </summary>
			int height;

			/// <summary>
			/// Initializes a new instance of the WindowSize class.
			/// </summary>
			/// <param name="width">the window width</param>
			/// <param name="height">the window height</param>
			public WindowSize(int width, int height)
			{
				this.width  = width;
				this.height = height;
			}

			/// <summary>
			/// Gets the window size.
			/// </summary>
			public Size ToSize()
			{
				return new Size(width, height);
			}

			/// <summary>
			/// Sets the window size.
			/// </summary>
			public void FromSize(Size size)
			{
				width  = size.Width;
				height = size.Height;
			}

			/// <summary>
			/// Gets or sets the window width.
			/// </summary>
			public int Width
			{
				get {
					return width;
				}
				set {
					width = value;
				}
			}

			/// <summary>
			/// Gets or sets the window height.
			/// </summary>
			public int Height
			{
				get {
					return height;
				}
				set {
					height = value;
				}
			}
		}

		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// The one and only Config instance (singleton).
		/// </summary>
		static Config instance = null;

		/// <summary>
		/// This property indicates whether logging is enabled or disabled.
		/// </summary>
		static bool loggingEnabled = false;

		/// <summary>
		/// The application window size.
		/// </summary>
		WindowSize clientSize = new WindowSize(644, 600);

		/// <summary>
		/// The window size of the script editor dialog.
		/// </summary>
		WindowSize scriptEditorSize = new WindowSize(600, 440);

		/// <summary>
		/// The window size of the text viewer dialog.
		/// </summary>
		WindowSize textViewerSize = new WindowSize(600, 440);

		/// <summary>
		/// The window size of the dictionary viewer dialog.
		/// </summary>
		WindowSize dictViewerSize = new WindowSize(600, 440);

		/// <summary>
		/// This property indicates whether the toolbar is visible or hidden.
		/// </summary>
		bool showToolbar = true;

		/// <summary>
		/// This property indicates whether the status bar is visible or hidden.
		/// </summary>
		bool showStatusBar = true;

		/// <summary>
		/// This property indicates whether the title bar is visible or hidden.
		/// </summary>
		bool showTitleBar = true;

		/// <summary>
		/// This property indicates whether the timer is visible or hidden.
		/// </summary>
		bool showTimer = true;

		/// <summary>
		/// This property indicates whether the script playback toolbar is
		/// visible or hidden.
		/// </summary>
		bool showScriptPlayback = true;

		/// <summary>
		/// The text color.
		/// </summary>
		Color textColor = Color.LightGray;

		/// <summary>
		/// The input text color.
		/// </summary>
		Color inputColor = Color.Yellow;

		/// <summary>
		/// The background color.
		/// </summary>
		Color backColor = Color.Black;

		/// <summary>
		/// The title/status bar text color.
		/// </summary>
		Color barTextColor = Color.Black;

		/// <summary>
		/// The title/status bar background color.
		/// </summary>
		Color barBackColor = Color.LightGray;

		/// <summary>
		/// The Amiga palette colors.
		/// </summary>
		Color[] amigaPalette =
		{
			Color.FromArgb(0x00,0x00,0x00),
			Color.FromArgb(0xff,0x00,0x00),
			Color.FromArgb(0x30,0xe8,0x30),
			Color.FromArgb(0xff,0xff,0x00),
			Color.FromArgb(0x00,0x00,0xff),
			Color.FromArgb(0xa0,0x68,0x00),
			Color.FromArgb(0x00,0xff,0xff),
			Color.FromArgb(0xff,0xff,0xff)
		};

		/// <summary>
		/// The Spectrum palette colors.
		/// </summary>
		Color[] spectrumPalette =
		{
			Color.FromArgb(0x00,0x00,0x00),
			Color.FromArgb(0x00,0x00,0xfc),
			Color.FromArgb(0x00,0xfc,0x00),
			Color.FromArgb(0x00,0xfc,0xfc),
			Color.FromArgb(0xfc,0x00,0x00),
			Color.FromArgb(0xf0,0x00,0xfc),
			Color.FromArgb(0xfc,0xfc,0x00),
			Color.FromArgb(0xfc,0xfc,0xfc),
		};

		/// <summary>
		/// The user defined palette colors.
		/// </summary>
		Color[] userPalette =
		{
			Color.FromArgb(0x00,0x00,0x00),
			Color.FromArgb(0xff,0x00,0x00),
			Color.FromArgb(0x30,0xe8,0x30),
			Color.FromArgb(0xff,0xff,0x00),
			Color.FromArgb(0x00,0x00,0xff),
			Color.FromArgb(0xa0,0x68,0x00),
			Color.FromArgb(0x00,0xff,0xff),
			Color.FromArgb(0xff,0xff,0xff)
		};

		/// <summary>
		/// The selected color palette for line drawn graphics.
		/// </summary>
		Palette paletteSelection = Palette.Amiga;

		/// <summary>
		/// The available image scaling factors for line drawn and bitmap
		/// graphics.
		/// </summary>
		int[] imageScalings =
		{
			1, 2, 3, 4
		};

		/// <summary>
		/// The selected image scaling factor.
		/// </summary>
		int imageScaling = 2;

		/// <summary>
		/// The in-game bitmap font.
		/// </summary>
		BitmapFont font = BitmapFonts.Get("Atari ST", 1);

		/// <summary>
		/// The selected font scaling factor.
		/// </summary>
		int fontScaling = 1;

		/// <summary>
		/// The available line spacings.
		/// </summary>
		int[] lineSpacings =
		{
			1, 2, 3, 4, 5, 6, 7, 8, 9, 10
		};

		/// <summary>
		/// The selected line spacing.
		/// </summary>
		int lineSpacing = 1;

		/// <summary>
		/// The list of 'recent files' in the main menu.
		/// </summary>
		string[] recentFiles;

		/// <summary>
		/// This property indicates whether or not Level9.Net scripts
		/// include comments.
		/// </summary>
		bool scriptIncludeComments = false;

		/// <summary>
		/// This property indicates whether Level9.Net scripts contain
		/// one or multiple entries per line.
		/// </summary>
		bool scriptOneEntryPerLine = false;

		/// <summary>
		/// The input delay for script playback (seconds).
		/// </summary>
		int scriptInputDelay = 2;

		/// <summary>
		/// The available script input delays.
		/// </summary>
		int[] scriptInputDelays =
		{
			0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
		};

		// -- Events ---------------------------------------------------------

		/// <summary>
		/// Occurs when the configuration has changed.
		/// </summary>
		public event ConfigChangedEventHandler ConfigChanged;

		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the Config class.
		/// </summary>
		private Config()
		{
			if (File.Exists(GetConfigFilePath())) {
				ReadSettings();
			}
		}

		// -- Methods --------------------------------------------------------

		/// <summary>
		/// Gets the current palette colors for line drawn graphics.
		/// </summary>
		public Color[] GetActivePalette()
		{
			Color[] palette = amigaPalette;

			switch (paletteSelection) {
				case Palette.Amiga:
					palette = amigaPalette;
					break;
				case Palette.Spectrum:
					palette = spectrumPalette;
					break;
				case Palette.UserDefined:
					palette = userPalette;
					break;
			}

			return palette;
		}

		/// <summary>
		/// Adds a file to the 'recent files' list.
		/// </summary>
		/// <param name="file">the file path path</param>
		public void AddToRecentFiles(string file)
		{
			if (recentFiles == null) {
				recentFiles = new string[MAX_RECENT_FILES];
			}
			if (recentFiles.Length < MAX_RECENT_FILES) {
				Array.Resize(ref recentFiles, MAX_RECENT_FILES);
			}
			int i = 0;
			bool addFile = (file != null);
			while (addFile && i < recentFiles.Length) {
				if (file.Equals(recentFiles[i++])) {
					addFile = false;
				}
			}
			if (addFile) {
				for (int j = recentFiles.Length - 1; j > 0; j--) {
					if (recentFiles[j - 1] != null) {
						recentFiles[j] = recentFiles[j - 1];
					}
				}
				recentFiles[0] = file;
			}
		}

		/// <summary>
		/// Notifies listeners that the configuration has changed.
		/// </summary>
		public void FireConfigChanged()
		{
			if (ConfigChanged != null) {
				ConfigChanged(new ConfigChangedEventArgs());
			}
		}

		/// <summary>
		/// Writes a XML settings file. The file path is given by
		/// GetConfigFilePath().
		/// </summary>
		public void WriteSettings()
		{
			TextWriter writer = null;

			try {
				XmlSerializer serializer = new XmlSerializer(typeof(ConfigXML));
				writer = new StreamWriter(GetConfigFilePath());
				serializer.Serialize(writer, new ConfigXML(this));
				writer.Close();
			}
			catch (Exception ex) {
				Logger.Log(this, "WriteSettingsXml", ex);
				if (writer != null) {
					writer.Close();
				}
			}
		}

		/// <summary>
		/// Reads a XML settings file. The file path is given by
		/// GetConfigFilePath().
		/// </summary>
		void ReadSettings()
		{
			FileStream stream = null;

			try {
				XmlSerializer serializer = new XmlSerializer(typeof(ConfigXML));
				stream = new FileStream(GetConfigFilePath(), FileMode.Open);
				XmlReader reader = new XmlTextReader(stream);
				ConfigXML settings = (ConfigXML)serializer.Deserialize(reader);
				SetMembers(settings);
				stream.Close();
			}
			catch (Exception ex) {
				ErrorHandler.Handle(this, "Error reading settings "
					+ "file - Using default configuration." , ex);
				if (stream != null) {
					stream.Close();
				}
			}
		}

		/// <summary>
		/// Checks the values of a ConfigXML instance and applies the settings
		/// to the current Config instance.
		/// </summary>
		/// <param name="xml">a ConfigXML instance</param>
		void SetMembers(ConfigXML xml)
		{
			if (xml == null) {
				return;
			}

			showStatusBar         = xml.show_status_bar;
			showToolbar           = xml.show_toolbar;
			showTitleBar          = xml.show_title_bar;
			showScriptPlayback    = xml.show_script_playback;
			showTimer             = xml.show_timer;

			scriptIncludeComments = xml.script_include_comments;
			scriptOneEntryPerLine = xml.script_one_entry_per_line;
			scriptInputDelay      = xml.script_input_delay;

			if ((xml.back_color & 0xff000000) == 0xff000000) {
				backColor = Color.FromArgb(xml.back_color);
			}

			if ((xml.text_color & 0xff000000) == 0xff000000
					&& xml.text_color != xml.back_color) {
				textColor = Color.FromArgb(xml.text_color);
			}

			if ((xml.input_color & 0xff000000) == 0xff000000
					&& xml.input_color != xml.back_color) {
				inputColor = Color.FromArgb(xml.input_color);
			}

			if ((xml.bar_back_color & 0xff000000) == 0xff000000) {
				barBackColor = Color.FromArgb(xml.bar_back_color);
			}

			if ((xml.bar_text_color & 0xff000000) == 0xff000000
					&& xml.bar_text_color != xml.bar_back_color) {
				barTextColor = Color.FromArgb(xml.bar_text_color);
			}

			if (BitmapFonts.Get(xml.font_type, xml.font_scaling) != null) {
				font        = BitmapFonts.Get(xml.font_type, xml.font_scaling);
				fontScaling = xml.font_scaling;
			}

			if (xml.line_spacing <= lineSpacings[lineSpacings.Length - 1]
					&& xml.line_spacing >= lineSpacings[0]) {
				lineSpacing = xml.line_spacing;
			}

			if (xml.image_scaling <= imageScalings[imageScalings.Length - 1]
					&& xml.image_scaling >= imageScalings[0]) {
				imageScaling = xml.image_scaling;
			}

			if (xml.recent_files != null) {
				recentFiles = xml.recent_files;
			}

			if (xml.palette_selection >= (int)Config.Palette.UserDefined
					&& xml.palette_selection <= (int)Config.Palette.Spectrum) {
				paletteSelection = (Config.Palette)xml.palette_selection;
			}

			if (xml.user_palette != null && xml.user_palette.Length > 0) {
				for (int i = 0; i < xml.user_palette.Length; i++) {
					userPalette[i] = Color.FromArgb(xml.user_palette[i]);
				}
			}

			if (xml.client_width > 0 && xml.client_height > 0) {
				clientSize.Width  = xml.client_width;
				clientSize.Height = xml.client_height;
			}

			if (xml.script_editor_width > 0 && xml.script_editor_height > 0) {
				scriptEditorSize.Width  = xml.script_editor_width;
				scriptEditorSize.Height = xml.script_editor_height;
			}

			if (xml.text_viewer_width > 0 && xml.text_viewer_height > 0) {
				textViewerSize.Width  = xml.text_viewer_width;
				textViewerSize.Height = xml.text_viewer_height;
			}

			if (xml.dict_viewer_width > 0 && xml.dict_viewer_height > 0) {
				dictViewerSize.Width  = xml.dict_viewer_width;
				dictViewerSize.Height = xml.dict_viewer_height;
			}
		}

		/// <summary>
		/// Provides the initialization state of the configuration.
		/// </summary>
		/// <returns>
		/// true, if the configuration has been initialized; otherwise, false
		/// </returns>
		public static bool IsInitialized()
		{
			return (instance != null);
		}

		/// <summary>
		/// Provides the application data path.
		/// </summary>
		/// <returns>the configuration file path</returns>
		public static string GetAppDataPath()
		{
			return Environment.GetFolderPath(
				Environment.SpecialFolder.ApplicationData) + "\\"
				+ PROGRAM_NAME;
		}

		/// <summary>
		/// Provides the configuration file path.
		/// </summary>
		/// <returns>the configuration file path</returns>
		public static string GetConfigFilePath()
		{
			return GetAppDataPath() + "\\" + SETTINGS_FILE;
		}

		/// <summary>
		/// Provides the Level 9 DLL path.
		/// </summary>
		/// <returns>the Level 9 DLL path</returns>
		public static string GetLevel9DLLPath()
		{
			return Application.StartupPath + "\\" + LEVEL9_DLL;
		}

		/// <summary>
		/// Provides the ANTLRv3 DLL path.
		/// </summary>
		/// <returns>the ANTLRv3 DLL path</returns>
		public static string GetAntlrPath()
		{
			return Application.StartupPath + "\\" + ANTLR_DLL;
		}

		/// <summary>
		/// Provides the L9cut executable path.
		/// </summary>
		/// <returns>the L9cut executable path</returns>
		public static string GetL9CutPath()
		{
			return Application.StartupPath + "\\" + L9CUT_EXE;
		}

		/// <summary>
		/// Provides the log file path.
		/// </summary>
		/// <returns>the log file path</returns>
		public static string GetLogFilePath()
		{
			return GetAppDataPath() + "\\" + LOG_FILE;
		}

		/// <summary>
		/// Provides the help file path.
		/// </summary>
		/// <returns>the help file path</returns>
		public static string GetHelpFilePath()
		{
			return Application.StartupPath + "\\" + HELP_FILE;
		}

		// -- Accessors ------------------------------------------------------

		/// <summary>
		/// Gets the one and only Config instance (singleton).
		/// </summary>
		public static Config Instance
		{
			get {
				if (instance == null) {
					instance = new Config();
				}
				return instance;
			}
		}

		/// <summary>
		/// Gets or sets the application window size.
		/// </summary>
		public WindowSize ClientSize
		{
			get {
				return clientSize;
			}
		}

		/// <summary>
		/// Gets or sets the window size of the script editor dialog.
		/// </summary>
		public WindowSize ScriptEditorSize
		{
			get {
				return scriptEditorSize;
			}
		}

		/// <summary>
		/// Gets or sets the window size of the text viewer dialog.
		/// </summary>
		public WindowSize TextViewerSize
		{
			get {
				return textViewerSize;
			}
		}

		/// <summary>
		/// Gets or sets the window size of the dictionary viewer dialog.
		/// </summary>
		public WindowSize DictViewerSize
		{
			get {
				return dictViewerSize;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether the toolbar is visible
		/// or hidden.
		/// </summary>
		public bool ShowToolbar
		{
			get {
				return showToolbar;
			}
			set {
				showToolbar = value;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether the status bar is
		/// visible or hidden.
		/// </summary>
		public bool ShowStatusBar
		{
			get {
				return showStatusBar;
			}
			set {
				showStatusBar = value;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether the title bar is
		/// visible or hidden.
		/// </summary>
		public bool ShowTitleBar
		{
			get {
				return showTitleBar;
			}
			set {
				showTitleBar = value;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether the script playback
		/// toolbar is visible or hidden.
		/// </summary>
		public bool ShowScriptPlayback
		{
			get {
				return showScriptPlayback;
			}
			set {
				showScriptPlayback = value;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether the timer is visible
		/// or hidden.
		/// </summary>
		public bool ShowTimer
		{
			get {
				return showTimer;
			}
			set {
				showTimer = value;
			}
		}

		/// <summary>
		/// Gets or sets the text color.
		/// </summary>
		public Color TextColor
		{
			get {
				return textColor;
			}
			set {
				textColor = value;
			}
		}

		/// <summary>
		/// Gets or sets the background color.
		/// </summary>
		public Color BackColor
		{
			get {
				return backColor;
			}
			set {
				backColor = value;
			}
		}

		/// <summary>
		/// Gets or sets the input text color.
		/// </summary>
		public Color InputColor
		{
			get {
				return inputColor;
			}
			set {
				inputColor = value;
			}
		}

		/// <summary>
		/// Gets or sets the title/status bar text color.
		/// </summary>
		public Color BarTextColor
		{
			get {
				return barTextColor;
			}
			set {
				barTextColor = value;
			}
		}

		/// <summary>
		/// Gets or sets the title/status bar background color.
		/// </summary>
		public Color BarBackColor
		{
			get {
				return barBackColor;
			}
			set {
				barBackColor = value;
			}
		}

		/// <summary>
		/// Gets the Amiga palette colors.
		/// </summary>
		public Color[] AmigaPalette
		{
			get {
				return amigaPalette;
			}
		}

		/// <summary>
		/// Gets the Spectrum palette colors.
		/// </summary>
		public Color[] SpectrumPalette
		{
			get {
				return spectrumPalette;
			}
		}

		/// <summary>
		/// Gets or sets the user defined palette colors.
		/// </summary>
		public Color[] UserPalette
		{
			get {
				return userPalette;
			}
			set {
				userPalette = value;
			}
		}

		/// <summary>
		/// Gets or sets the selected color palette for line drawn graphics.
		/// </summary>
		public Palette PaletteSelection
		{
			get {
				return paletteSelection;
			}
			set {
				paletteSelection = value;
			}
		}

		/// <summary>
		/// Gets or sets the in-game bitmap font.
		/// </summary>
		public BitmapFont Font
		{
			get {
				return font;
			}
			set {
				font = value;
			}
		}

		/// <summary>
		/// Gets or sets the selected font scaling factor.
		/// </summary>
		public int FontScaling
		{
			get {
				return fontScaling;
			}
			set {
				fontScaling = value;
			}
		}

		/// <summary>
		/// Gets or sets the selected lines spacing.
		/// </summary>
		public int LineSpacing
		{
			get {
				return lineSpacing;
			}
			set {
				lineSpacing = value;
			}
		}

		/// <summary>
		/// Gets the available line spacings.
		/// </summary>
		public int[] LineSpacings
		{
			get {
				return lineSpacings;
			}
		}

		/// <summary>
		/// Gets or sets the selected image scaling factor.
		/// </summary>
		public int ImageScaling
		{
			get {
				return imageScaling;
			}
			set {
				imageScaling = value;
			}
		}

		/// <summary>
		/// Gets the available image scaling factors for line drawn and bitmap
		/// graphics.
		/// </summary>
		public int[] ImageScalings
		{
			get {
				return imageScalings;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether or not Level9.Net
		/// scripts include comments.
		/// </summary>
		public bool ScriptIncludeComments
		{
			get {
				return scriptIncludeComments;
			}
			set {
				scriptIncludeComments = value;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether Level9.Net scripts
		/// contain one or multiple entries per line.
		/// </summary>
		public bool ScriptOneEntryPerLine
		{
			get {
				return scriptOneEntryPerLine;
			}
			set {
				scriptOneEntryPerLine = value;
			}
		}

		/// <summary>
		/// Gets or sets the input delay for script playback (seconds).
		/// </summary>
		public int ScriptInputDelay
		{
			get {
				return scriptInputDelay;
			}
			set {
				scriptInputDelay = value;
			}
		}

		/// <summary>
		/// Gets the available script playback delays.
		/// </summary>
		public int[] ScriptInputDelays
		{
			get {
				return scriptInputDelays;
			}
		}

		/// <summary>
		/// Gets or sets the list of 'recent files' in the main menu.
		/// </summary>
		public string[] RecentFiles
		{
			get {
				return recentFiles;
			}
			set {
				recentFiles = value;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether logging is enabled or
		/// disabled.
		/// </summary>
		public static bool LoggingEnabled
		{
			get {
				return loggingEnabled;
			}
			set {
				loggingEnabled = value;
			}
		}
	}

	/// <summary>
	/// Allows XML serialization of miscellaneous configuration options.
	/// </summary>
	public class ConfigXML
	{
		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// See <see cref="Config.clientSize"></see>.
		/// </summary>
		public int client_width;

		/// <summary>
		/// See <see cref="Config.clientSize"></see>.
		/// </summary>
		public int client_height;

		/// <summary>
		/// See <see cref="Config.scriptEditorSize"></see>.
		/// </summary>
		public int script_editor_width;

		/// <summary>
		/// See <see cref="Config.scriptEditorSize"></see>.
		/// </summary>
		public int script_editor_height;

		/// <summary>
		/// See <see cref="Config.textViewerSize"></see>.
		/// </summary>
		public int text_viewer_width;

		/// <summary>
		/// See <see cref="Config.textViewerSize"></see>.
		/// </summary>
		public int text_viewer_height;

		/// <summary>
		/// See <see cref="Config.dictViewerSize"></see>.
		/// </summary>
		public int dict_viewer_width;

		/// <summary>
		/// See <see cref="Config.dictViewerSize"></see>.
		/// </summary>
		public int dict_viewer_height;

		/// <summary>
		/// See <see cref="Config.showToolbar"></see>.
		/// </summary>
		public bool show_toolbar;

		/// <summary>
		/// See <see cref="Config.showStatusBar"></see>.
		/// </summary>
		public bool show_status_bar;

		/// <summary>
		/// See <see cref="Config.showScriptPlayback"></see>.
		/// </summary>
		public bool show_script_playback;

		/// <summary>
		/// See <see cref="Config.showTimer"></see>.
		/// </summary>
		public bool show_timer;

		/// <summary>
		/// See <see cref="Config.showTitleBar"></see>.
		/// </summary>
		public bool show_title_bar;

		/// <summary>
		/// See <see cref="Config.backColor"></see>.
		/// </summary>
		public int back_color;

		/// <summary>
		/// See <see cref="Config.textColor"></see>.
		/// </summary>
		public int text_color;

		/// <summary>
		/// See <see cref="Config.inputColor"></see>.
		/// </summary>
		public int input_color;

		/// <summary>
		/// See <see cref="Config.barTextColor"></see>.
		/// </summary>
		public int bar_text_color;

		/// <summary>
		/// See <see cref="Config.barBackColor"></see>.
		/// </summary>
		public int bar_back_color;

		/// <summary>
		/// See <see cref="Config.userPalette"></see>.
		/// </summary>
		public int[] user_palette;

		/// <summary>
		/// See <see cref="Config.font"></see>.
		/// </summary>
		public string font_type;

		/// <summary>
		/// See <see cref="Config.fontScaling"></see>.
		/// </summary>
		public int font_scaling;

		/// <summary>
		/// See <see cref="Config.lineSpacing"></see>.
		/// </summary>
		public int line_spacing;

		/// <summary>
		/// See <see cref="Config.imageScaling"></see>.
		/// </summary>
		public int image_scaling;

		/// <summary>
		/// See <see cref="Config.paletteSelection"></see>.
		/// </summary>
		public int palette_selection;

		/// <summary>
		/// See <see cref="Config.recentFiles"></see>.
		/// </summary>
		public string[] recent_files;

		/// <summary>
		/// See <see cref="Config.scriptIncludeComments"></see>.
		/// </summary>
		public bool script_include_comments;

		/// <summary>
		/// See <see cref="Config.scriptOneEntryPerLine"></see>.
		/// </summary>
		public bool script_one_entry_per_line;

		/// <summary>
		/// See <see cref="Config.scriptInputDelay"></see>.
		/// </summary>
		public int script_input_delay;

		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Parameterless constructor for serialization.
		/// </summary>
		public ConfigXML() {}

		/// <summary>
		/// Initializes a new instance of the ConfigXML class.
		/// </summary>
		/// <param name="config">a Config instance</param>
		public ConfigXML(Config config)
		{
			client_height        = config.ClientSize.Height;
			client_width         = config.ClientSize.Width;
			script_editor_width  = config.ScriptEditorSize.Width;
			script_editor_height = config.ScriptEditorSize.Height;
			text_viewer_width    = config.TextViewerSize.Width;
			text_viewer_height   = config.TextViewerSize.Height;
			dict_viewer_width    = config.DictViewerSize.Width;
			dict_viewer_height   = config.DictViewerSize.Height;
			show_status_bar      = config.ShowStatusBar;
			show_toolbar         = config.ShowToolbar;
			show_title_bar       = config.ShowTitleBar;
			show_script_playback = config.ShowScriptPlayback;
			show_timer           = config.ShowTimer;
			back_color           = config.BackColor.ToArgb();
			text_color           = config.TextColor.ToArgb();
			input_color          = config.InputColor.ToArgb();
			bar_text_color       = config.BarTextColor.ToArgb();
			bar_back_color       = config.BarBackColor.ToArgb();
			input_color          = config.InputColor.ToArgb();
			font_type            = config.Font.Type;
			font_scaling         = config.FontScaling;
			line_spacing         = config.LineSpacing;
			image_scaling        = config.ImageScaling;
			recent_files         = config.RecentFiles;

			palette_selection = (int)config.PaletteSelection;
			user_palette = new int[config.UserPalette.Length];
			for (int i = 0; i < config.UserPalette.Length; i++) {
				user_palette[i] = config.UserPalette[i].ToArgb();
			}

			script_include_comments   = config.ScriptIncludeComments;
			script_one_entry_per_line = config.ScriptOneEntryPerLine;
			script_input_delay        = config.ScriptInputDelay;
		}
	}

	/// <summary>
	/// Represents the method that handles a ConfigChanged event.
	/// </summary>
	public delegate void ConfigChangedEventHandler(ConfigChangedEventArgs e);

	/// <summary>
	/// Provides data for the ConfigChanged event.
	/// </summary>
	public class ConfigChangedEventArgs : EventArgs
	{
		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the ConfigChangedEventArgs class.
		/// </summary>
		public ConfigChangedEventArgs() { }
	}
}
