/*
 * DictionaryProgressForm.cs - Displays a progress bar while reading
 *                             the dictionary of a Level 9 game.
 *
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

using System;
using System.Windows.Forms;

namespace Level9
{
	/// <summary>
	/// Displays a progress bar while reading the dictionary of a Level 9 game.
	/// </summary>
	public partial class DictionaryProgressForm : Form
	{
		// -- Constants ------------------------------------------------------

		/// <summary>
		/// The maximum time for a dictionary import in milliseconds.
		/// </summary>
		public const int WATCHDOG_TIMEOUT = 20000;

		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// A watchdog timer to control the dictionary import.
		/// </summary>
		System.Windows.Forms.Timer watchdogTimer =
			new System.Windows.Forms.Timer();

		/// <summary>
		/// A handler for OsFlush events raised by the interpreter.
		/// </summary>
		OsFlushEventHandler osFlushHandler;

		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the DictionaryProgressForm class.
		/// </summary>
		public DictionaryProgressForm()
		{
			InitializeComponent();
			progressBarRead.Value = progressBarRead.Minimum;
			osFlushHandler = new OsFlushEventHandler(InterpreterOsFlush);
			Interpreter.OsFlush += osFlushHandler;
			watchdogTimer.Tick += new EventHandler(WatchdogTimerTick);
			watchdogTimer.Interval = WATCHDOG_TIMEOUT;
		}

		// -- Event handling -------------------------------------------------

		/// <summary>
		/// Invoked when the interpreter flushes a portion of text.
		/// </summary>
		/// <param name="e">
		/// an OsFlushEventArgs that contains the event data
		/// </param>
		void InterpreterOsFlush(OsFlushEventArgs e)
		{
			if (this.InvokeRequired == false
					&& e.Command == Interpreter.Commands.Dictionary) {
				watchdogTimer.Start();
				Dictionary.Instance.Add(e.Text);
				if (progressBarRead.Value == progressBarRead.Maximum) {
					progressBarRead.Value = progressBarRead.Minimum;
				}
				progressBarRead.PerformStep();
				if (e.Text.EndsWith("\0")) {
					Interpreter.OsFlush -= osFlushHandler;
					this.DialogResult = DialogResult.OK;
					this.Close();
				}
			}
			else {
				OsFlushEventHandler osFlushHandler =
					new OsFlushEventHandler(InterpreterOsFlush);
				this.Invoke(osFlushHandler, new object[] { e });
			}
		}

		/// <summary>
		/// Invoked when a watchdog timer interval has expired.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void WatchdogTimerTick(Object sender, EventArgs e)
		{
			Interpreter.OsFlush -= osFlushHandler;
			this.DialogResult = DialogResult.OK;
			this.Close();
		}
	}
}
