/*
 * Logger.cs - Simple logging functionality for Level9.Net.
 *
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

using System;
using System.IO;

namespace Level9
{
	/// <summary>
	/// Simple logging functionality for Level9.Net
	/// </summary>
	public class Logger
	{
		// -- Constants ------------------------------------------------------

		/// <summary>
		/// The format of logging timestamps.
		/// </summary>
		const string TS_FORMAT = "yyyy-MM-dd HH:mm:ss";

		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// A stream writer for writing into a logfile.
		/// </summary>
		private static StreamWriter sw = null;

		// -- Methods --------------------------------------------------------

		/// <summary>
		/// Appends a log message to the application logfile.
		/// </summary>
		/// <param name="type">
		/// the object type that raises the log message
		/// </param>
		/// <param name="message">the log message</param>
		/// <param name="e">an Exception, or null if none occurred</param>
		public static void Log(Type type, string message, Exception e)
		{
			if (!Config.LoggingEnabled) {
				return;
			}

			try {
				if (sw == null) {
					sw = File.AppendText(Config.GetLogFilePath());
				}
				lock(sw) {
					sw.WriteLine(DateTime.Now.ToString(TS_FORMAT)
						+ " [" + type + "] " + message);
					if (e != null) {
						sw.WriteLine(e.Message);
						sw.WriteLine(e.StackTrace);
					}
					sw.Flush();
				}
			}
			catch  {
				// Do nothing
			}
		}

		/// <summary>
		/// Appends a log message to the application logfile.
		/// </summary>
		/// <param name="obj">the object that raises the log message</param>
		/// <param name="message">the log message</param>
		/// <param name="e">an Exception, or null if none occurred</param>
		public static void Log(Object obj, string message, Exception e)
		{
			if (obj != null) {
				Log(obj.GetType(), message, e);
			}
		}

		/// <summary>
		/// Appends a log message to the application logfile.
		/// </summary>
		/// <param name="type">
		/// the object type that raises the log message
		/// </param>
		/// <param name="message">the log message</param>
		public static void Log(Type type, string message)
		{
			Log(type, message, null);
		}

		/// <summary>
		/// Appends a log message to the application logfile.
		/// </summary>
		/// <param name="obj">the object that raises the log message</param>
		/// <param name="message">the log message</param>
		public static void Log(Object obj, string message)
		{
			Log(obj, message, null);
		}
	}
}
