/*
 * SimpleListView.cs - Simple list view control with icons and captions.
 *
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace Level9
{
	/// <summary>
	/// Simple list view control with icons and captions.
	/// </summary>
	public class SimpleListView : Panel
	{
		// -- Constants ------------------------------------------------------

		/// <summary>
		/// The vertical spacing between list view items in pixels.
		/// </summary>
		const int V_SPACING = 7;

		/// <summary>
		/// The spacing between component border and icons in pixels.
		/// </summary>
		const int BORDER_SPACING  = 5;

		/// <summary>
		/// The vertical spacing between icons and captions in pixels.
		/// </summary>
		const int CAPTION_SPACING = 3;

		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// A list of list view items.
		/// </summary>
		List<SimpleListViewItem> listViewItems =
			new List<SimpleListViewItem>();

		/// <summary>
		/// The number of the currently selected item. A value of -1 means that
		/// no item is selected.
		/// </summary>
		int selected = -1;

		// -- Events ---------------------------------------------------------

		/// <summary>Occurs when a list view item has been clicked.</summary>
		public event ListViewItemClickedEventHandler ItemClicked;

		// -- Constructor ----------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the SimpleListView class.
		/// </summary>
		public SimpleListView()
		{
			this.SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
			this.SetStyle(ControlStyles.UserPaint, true);
			this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			this.SetStyle(ControlStyles.ResizeRedraw, true);
		}

		// -- Event handling -------------------------------------------------

		/// <summary>
		/// Handles the MouseDown event.
		/// </summary>
		protected override void OnMouseDown(MouseEventArgs e)
		{
			base.OnMouseDown(e);
			int item = GetItemByPoint(new Point(e.X, e.Y));
			if (item >= 0) {
				int prevSelected = selected;
				selected = item;
				if (ItemClicked != null) {
					ListViewItemClickedEventArgs args = new
						ListViewItemClickedEventArgs(listViewItems[item]);
					ItemClicked(this, args);
				}
				InvalidateItem(selected);
				if (prevSelected >= 0) {
					InvalidateItem(prevSelected);
				}
			}
		}

		/// <summary>
		/// Handles the Paint event.
		/// </summary>
		protected override void OnPaint(PaintEventArgs e)
		{
			base.OnPaint(e);
			Graphics g = e.Graphics;

			SolidBrush fgBrush = new SolidBrush(this.ForeColor);
			SolidBrush bgBrush = new SolidBrush(this.BackColor);
			SolidBrush textBrush = null;

			for (int i = 0; i < listViewItems.Count; i++) {
				SimpleListViewItem item = listViewItems[i];
				Rectangle rect = GetItemRect(i);
				if (!rect.IntersectsWith(e.ClipRectangle)) {
					continue;
				}
				//
				// Draw background
				//
				if (i == selected) {
					g.FillRectangle(SystemBrushes.ActiveCaption, rect);
					textBrush = bgBrush;
				}
				else {
					g.FillRectangle(bgBrush, rect);
					textBrush = fgBrush;
				}
				//
				// Draw image
				//
				int x = this.Width / 2 - item.Image.Width / 2
					- SystemInformation.Border3DSize.Width;
				int y = rect.Top + BORDER_SPACING;
				g.DrawImage(item.Image, x, y, item.Image.Width,
					item.Image.Height);
				//
				// Draw caption
				//
				x = this.Width / 2 - GetCaptionWidth(item.Text) / 2
					- SystemInformation.Border3DSize.Width;
				y = rect.Top + BORDER_SPACING + item.Image.Height
					+ CAPTION_SPACING;
				g.DrawString(item.Text, this.Font, textBrush, x, y);
			}

			fgBrush.Dispose();
			bgBrush.Dispose();
		}

		// -- Methods --------------------------------------------------------

		/// <summary>
		/// Adds an item to the list view.
		/// </summary>
		/// <param name="listViewItem">a list view item</param>
		public void AddListViewItem(SimpleListViewItem listViewItem)
		{
			if (listViewItem != null && listViewItem.Image != null) {
				listViewItems.Add(listViewItem);
			}
		}

		/// <summary>
		/// Sets the selected list view item.
		/// </summary>
		/// <param name="listViewItem">the selected list view item</param>
		public void SetSelectedItem(SimpleListViewItem listViewItem)
		{
			if (listViewItem == null) {
				return;
			}

			for (int i = 0; i < listViewItems.Count; i++) {
				if (listViewItem != listViewItems[i]) {
					continue;
				}
				selected = i;
				if (ItemClicked != null) {
					ListViewItemClickedEventArgs args = new
						ListViewItemClickedEventArgs(listViewItems[i]);
					ItemClicked(this, args);
				}
				this.Invalidate();
				break;
			}
		}

		/// <summary>
		/// Invalidates a list view item that must be redrawn.
		/// </summary>
		/// <param name="item">a list view item</param>
		void InvalidateItem(int item)
		{
			this.Invalidate(GetItemRect(item));
		}

		/// <summary>
		/// Provides the list view item that's associated with a given point
		/// on the component.
		/// </summary>
		/// <param name="point">a point on the component</param>
		/// <returns>the list view item number</returns>
		int GetItemByPoint(Point point)
		{
			for (int i = 0; i < listViewItems.Count; i++) {
				Rectangle itemRect = GetItemRect(i);
				if (itemRect.Contains(point)) {
					return i;
				}
			}
			return -1;
		}

		/// <summary>
		/// Provides the rectangle that's associated with a list view item.
		/// </summary>
		/// <param name="item">the list view item number</param>
		/// <returns>a rectangle</returns>
		Rectangle GetItemRect(int item)
		{
			SimpleListViewItem listViewItem = null;
			int yOffset = V_SPACING;

			for (int i = 0; i < item; i++) {
				listViewItem = listViewItems[i];
				yOffset += V_SPACING + BORDER_SPACING * 2
					+ listViewItem.Image.Height + CAPTION_SPACING
					+ this.Font.Height;
			}
			listViewItem = listViewItems[item];
			int height = listViewItem.Image.Height + CAPTION_SPACING
				+ this.Font.Height + BORDER_SPACING * 2;
			int width = Math.Max(GetCaptionWidth(listViewItem.Text),
				listViewItem.Image.Width) + BORDER_SPACING * 2
				- SystemInformation.Border3DSize.Width;
			int xOffset = this.Width / 2 - width / 2 - 2;

			return new Rectangle(xOffset, yOffset, width, height);
		}

		/// <summary>
		/// Provides the width of a list view item's caption.
		/// </summary>
		/// <param name="caption">the caption</param>
		/// <returns>the caption width</returns>
		int GetCaptionWidth(string caption)
		{
			Graphics g = this.CreateGraphics();
			SizeF width = g.MeasureString(caption, this.Font);
			g.Dispose();

			return (int)Math.Ceiling(width.Width);
		}
	}

	/// <summary>
	/// Represents the method that handles a ListViewItemClicked event.
	/// </summary>
	public delegate void ListViewItemClickedEventHandler(Object sender,
		ListViewItemClickedEventArgs e);

	/// <summary>
	/// Provides data for the ListViewItemClicked event.
	/// </summary>
	public class ListViewItemClickedEventArgs : EventArgs
	{
		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// The list view item.
		/// </summary>
		SimpleListViewItem item = null;

		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the ListViewItemClickedEventArgs
		/// class.
		/// </summary>
		/// <param name="item">the list view item</param>
		public ListViewItemClickedEventArgs(SimpleListViewItem item)
		{
			this.item = item;
		}

		// -- Accessors ------------------------------------------------------

		/// <summary>
		/// Gets or sets the list view item.
		/// </summary>
		public SimpleListViewItem Item
		{
			get {
				return item;
			}
			set {
				item = value;
			}
		}
	}

	/// <summary>
	/// Represents a simple list view item.
	/// </summary>
	public class SimpleListViewItem
	{
		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// The list view item's image.
		/// </summary>
		Image image = null;

		/// <summary>
		/// The text of the list view item's caption.
		/// </summary>
		string text = null;

		// -- Constructor ----------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the SimpleListViewItem class.
		/// </summary>
		/// <param name="image">the list view item's image</param>
		/// <param name="text">the text of the list view item's caption</param>
		public SimpleListViewItem(Image image, string text)
		{
			this.image = image;
			this.text  = text;
		}

		// -- Accessors ------------------------------------------------------

		/// <summary>
		/// Gets or sets the list view item's image.
		/// </summary>
		public Image Image
		{
			get {
				return this.image;
			}
			set {
				this.image = value;
			}
		}

		/// <summary>
		/// Gets or sets the text of the list view item's caption.
		/// </summary>
		public string Text
		{
			get {
				return this.text;
			}
			set {
				this.text = value;
			}
		}
	}
}
