/*
 * SimpleToolBar.cs - Simple toolbar with icons and tooltips.
 *
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace Level9
{
	/// <summary>
	/// Simple toolbar with icons and tooltips.
	/// </summary>
	public class SimpleToolBar : Panel
	{
		// -- Constants ------------------------------------------------------

		/// <summary>
		/// The horizontal spacing between icons in pixels.
		/// </summary>
		const int H_SPACING = 6;

		/// <summary>
		/// The vertical spacing between component border and icons in pixels.
		/// </summary>
		const int V_SPACING = 4;

		/// <summary>
		/// The horizontal spacing between component border and icons in pixels.
		/// </summary>
		const int H_OFFSET  = 6;

		/// <summary>
		/// The icon border spacing in pixels.
		/// </summary>
		const int BORDER_SPACING  = 3;

		/// <summary>
		/// The width of a toolbar separator in pixels.
		/// </summary>
		const int SEPARATOR_WIDTH = 2;

		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// A list of toolbar buttons.
		/// </summary>
		List<SimpleToolBarButton> toolbarButtons =
			new List<SimpleToolBarButton>();

		/// <summary>
		/// The number of the currently selected button. A value of -1 means
		/// that no button is selected.
		/// </summary>
		int selected = -1;

		/// <summary>
		/// The number of the currently pressed button. A value of -1 means that
		/// no button is pressed.
		/// </summary>
		int pressed  = -1;

		/// <summary>
		/// The maximum height of a toolbar icon in pixels.
		/// </summary>
		int maxIconHeight = 0;

		/// <summary>
		/// The toolbar's tooltip.
		/// </summary>
		ToolTip toolTip = new ToolTip();

		/// <summary>
		/// A grey scale matrix to display disabled toolbar buttons.
		/// </summary>
		private static float[][] greyScaleMatrix = new float[][] {
			new float[] { 0.2222f, 0.2222f, 0.2222f, 0.0000f, 0.0000f },
			new float[] { 0.2222f, 0.2222f, 0.2222f, 0.0000f, 0.0000f },
			new float[] { 0.2222f, 0.2222f, 0.2222f, 0.0000f, 0.0000f },
			new float[] { 0.3333f, 0.3333f, 0.3333f, 0.7500f, 0.0000f },
			new float[] { 0.0000f, 0.0000f, 0.0000f, 0.0000f, 1.0000f }
		};

		// -- Events ---------------------------------------------------------

		/// <summary>Occurs when a toolbar button has been clicked.</summary>
		public event ToolbarButtonClickedEventHandler ButtonClicked;

		// -- Constructor ----------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the SimpleToolBar class.
		/// </summary>
		public SimpleToolBar()
		{
			this.SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
			this.SetStyle(ControlStyles.UserPaint, true);
			this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			this.SetStyle(ControlStyles.ResizeRedraw, true);
		}

		// -- Event handling -------------------------------------------------

		/// <summary>
		/// Handles the MouseMove event.
		/// </summary>
		protected override void OnMouseMove(MouseEventArgs e)
		{
			base.OnMouseMove(e);
			int button = GetButtonByPoint(new Point(e.X, e.Y));

			if (button >= 0) {
				if (button != selected) {
					if (selected >= 0) {
						int prevSelected = selected;
						selected = button;
						InvalidateButton(prevSelected);
						InvalidateButton(selected);
					}
					else {
						selected = button;
						InvalidateButton(selected);
					}
					SimpleToolBarButton toolbarButton = toolbarButtons[button];
					toolTip.SetToolTip(this, toolbarButton.ToolTip);
				}
			}
			else {
				if (pressed != -1) {
					InvalidateButton(pressed);
				}
				else if (selected != -1) {
					InvalidateButton(selected);
				}
				selected = -1;
				pressed  = -1;
				toolTip.SetToolTip(this, null);
			}
		}

		/// <summary>
		/// Handles the MouseDown event.
		/// </summary>
		protected override void OnMouseDown(MouseEventArgs e)
		{
			base.OnMouseDown(e);
			int button = GetButtonByPoint(new Point(e.X, e.Y));

			if (button >= 0) {
				pressed = button;
				InvalidateButton(button);
			}
		}

		/// <summary>
		/// Handles the MouseUp event.
		/// </summary>
		protected override void OnMouseUp(MouseEventArgs e)
		{
			base.OnMouseUp(e);
			int button = GetButtonByPoint(new Point(e.X, e.Y));

			if (button >= 0) {
				pressed = -1;
				SimpleToolBarButton toolbarButton = toolbarButtons[button];
				if (ButtonClicked != null && toolbarButton.Enabled) {
					ToolbarButtonClickedEventArgs args =
						new ToolbarButtonClickedEventArgs(
							toolbarButtons[button]);
					ButtonClicked(this, args);
				}
				InvalidateButton(button);
			}
		}

		/// <summary>
		/// Handles the MouseLeave event.
		/// </summary>
		protected override void OnMouseLeave(EventArgs e)
		{
			base.OnMouseLeave(e);

			if (pressed != -1) {
				InvalidateButton(pressed);
			}
			else if (selected != -1) {
				InvalidateButton(selected);
			}
			selected = -1;
			pressed  = -1;
		}

		/// <summary>
		/// Handles the Paint event.
		/// </summary>
		protected override void OnPaint(PaintEventArgs e)
		{
			base.OnPaint(e);
			Graphics g = e.Graphics;

			//
			// Line
			//
			g.DrawLine(SystemPens.ControlDark, e.ClipRectangle.Left, 0,
				e.ClipRectangle.Right, 0);
			g.DrawLine(SystemPens.ControlLightLight, e.ClipRectangle.Left, 1,
				e.ClipRectangle.Right, 1);

			//
			// Buttons
			//
			for (int i = 0; i < toolbarButtons.Count; i++) {
				SimpleToolBarButton button = toolbarButtons[i];
				Rectangle rect = GetButtonRect(i);
				if (!rect.IntersectsWith(e.ClipRectangle)) {
					continue;
				}
				if (button.IsSeparator) {
					g.DrawLine(SystemPens.ControlDark, rect.Left, rect.Top,
						rect.Left, rect.Bottom);
					g.DrawLine(SystemPens.ControlLightLight, rect.Left + 1,
						rect.Top, rect.Left + 1, rect.Bottom);
				}
				else {
					if (button.Enabled) {
						g.DrawImage(button.Image, rect.Left
							+ BORDER_SPACING, rect.Top + BORDER_SPACING,
							button.Image.Width, button.Image.Height);
					}
					else {
						Rectangle imageRect = new Rectangle(
							rect.Left + BORDER_SPACING,
							rect.Top + BORDER_SPACING,
							button.Image.Width, button.Image.Height);
						ImageAttributes greyScale = new ImageAttributes();
						greyScale.SetColorMatrix(
							new ColorMatrix(greyScaleMatrix));
						g.DrawImage(button.Image, imageRect, 0, 0,
							button.Image.Width, button.Image.Height,
							GraphicsUnit.Pixel, greyScale);
					}
					if ((i == pressed || i == selected) && button.Enabled) {
						Pen pen1, pen2;
						if (i == pressed) {
							pen1 = SystemPens.ControlDark;
							pen2 = SystemPens.ControlLightLight;
						}
						else {
							pen1 = SystemPens.ControlLightLight;
							pen2 = SystemPens.ControlDark;
						}
						g.DrawLine(pen1, rect.Left, rect.Top,
							rect.Left, rect.Bottom - 1);
						g.DrawLine(pen1, rect.Left, rect.Top,
							rect.Right - 1, rect.Top);
						g.DrawLine(pen2, rect.Right - 1, rect.Top,
							rect.Right - 1, rect.Bottom - 1);
						g.DrawLine(pen2, rect.Left, rect.Bottom - 1,
							rect.Right - 1, rect.Bottom - 1);
					}
				}
			}
		}

		/// <summary>
		/// Invoked when the enabled state of a toolbar button has changed.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void ToolbarButtonEnabledChanged(SimpleToolBarButton sender,
			EventArgs e)
		{
			InvalidateButton(toolbarButtons.IndexOf(sender));
		}

		// -- Methods --------------------------------------------------------

		/// <summary>
		/// Adds a toolbar button to the toolbar.
		/// </summary>
		/// <param name="toolbarButton">a toolbar button</param>
		public void AddButton(SimpleToolBarButton toolbarButton)
		{
			bool addButton = toolbarButton != null
				&& (toolbarButton.Image != null || toolbarButton.IsSeparator);

			if (addButton) {
				toolbarButtons.Add(toolbarButton);
				toolbarButton.EnabledChanged +=
					new ToolbarButtonEnabledChangedEventHandler(
						ToolbarButtonEnabledChanged);
				if (toolbarButton.Image != null
						&& toolbarButton.Image.Height > maxIconHeight) {
					maxIconHeight = toolbarButton.Image.Height;
					this.Height = V_SPACING * 2 + BORDER_SPACING * 2
						+ maxIconHeight;
				}
			}
		}

		/// <summary>
		/// Invalidates a toolbar button that must be redrawn.
		/// </summary>
		/// <param name="button">a toolbar button</param>
		void InvalidateButton(int button)
		{
			this.Invalidate(GetButtonRect(button));
		}

		/// <summary>
		/// Provides the button number that's associated with a point on the
		/// toolbar.
		/// </summary>
		/// <param name="point">a point on the toolbar</param>
		/// <returns>the button number</returns>
		int GetButtonByPoint(Point point)
		{
			for (int i = 0; i < toolbarButtons.Count; i++) {
				Rectangle buttonRect = GetButtonRect(i);
				if (buttonRect.Contains(point)) {
					return i;
				}
			}

			return -1;
		}

		/// <summary>
		/// Provides the rectangle that's associated with a toolbar button.
		/// </summary>
		/// <param name="button">the button number</param>
		/// <returns>a rectangle</returns>
		Rectangle GetButtonRect(int button)
		{
			SimpleToolBarButton toolbarButton = null;
			int offset = H_OFFSET;

			for (int i = 0; i < button; i++) {
				toolbarButton = toolbarButtons[i];
				if (toolbarButton.IsSeparator) {
					offset += H_SPACING + SEPARATOR_WIDTH;
				}
				else {
					offset += H_SPACING + BORDER_SPACING * 2
						+ toolbarButton.Image.Width;
				}
			}
			toolbarButton = toolbarButtons[button];
			int height = this.Height - 2 * V_SPACING;
			int width = toolbarButton.IsSeparator ? SEPARATOR_WIDTH
				: toolbarButton.Image.Width + 2 * BORDER_SPACING;

			return new Rectangle(offset, V_SPACING, width, height);
		}
	}

	/// <summary>
	/// Represents the method that handles a ToolbarButtonClicked event.
	/// </summary>
	public delegate void ToolbarButtonClickedEventHandler(Object sender,
		ToolbarButtonClickedEventArgs e);

	/// <summary>
	/// Provides data for the ToolbarButtonClicked event.
	/// </summary>
	public class ToolbarButtonClickedEventArgs : EventArgs
	{
		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// The toolbar button.
		/// </summary>
		SimpleToolBarButton button = null;

		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the ToolbarButtonClickedEventArgs
		/// class.
		/// </summary>
		/// <param name="button">a toolbar button</param>
		public ToolbarButtonClickedEventArgs(SimpleToolBarButton button)
		{
			this.button = button;
		}

		// -- Accessors ------------------------------------------------------

		/// <summary>
		/// Gets or sets the toolbar button.
		/// </summary>
		public SimpleToolBarButton Button
		{
			get {
				return button;
			}
			set {
				button = value;
			}
		}
	}

	/// <summary>
	/// Represents the method that handles a ToolbarButtonEnabledChanged event.
	/// </summary>
	public delegate void ToolbarButtonEnabledChangedEventHandler(
		SimpleToolBarButton sender, EventArgs e);

	/// <summary>
	/// Represents a simple toolbar button.
	/// </summary>
	public class SimpleToolBarButton
	{
		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// The toolbar button's image.
		/// </summary>
		Image image = null;

		/// <summary>
		/// This property indicates whether the toolbar button acts as a
		/// separator.
		/// </summary>
		bool separator = false;

		/// <summary>
		/// This property indicates whether the toolbar button is enabled.
		/// </summary>
		bool enabled = true;

		/// <summary>
		/// The text of the toolbar button's tooltip.
		/// </summary>
		string toolTip = null;

		// -- Events ---------------------------------------------------------

		/// <summary>
		/// Occurs when the enabled state of a toolbar button has changed.
		/// </summary>
		public event ToolbarButtonEnabledChangedEventHandler EnabledChanged;

		// -- Constructor ----------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the SimpleToolBarButton class.
		/// </summary>
		/// <param name="image">the toolbar button's image</param>
		/// <param name="toolTip">
		/// the text of the toolbar button's tooltip
		/// </param>
		public SimpleToolBarButton(Image image, string toolTip)
		{
			this.image = image;
			this.toolTip = toolTip;
		}

		/// <summary>
		/// Initializes a new instance of the SimpleToolBarButton class.
		/// This constructor creates toolbar button separators.
		/// </summary>
		private SimpleToolBarButton()
		{
			separator = true;
		}

		// -- Accessors ------------------------------------------------------

		/// <summary>
		/// Gets or sets a value that indicates whether the toolbar button
		/// is enabled.
		/// </summary>
		public bool Enabled
		{
			get {
				return this.enabled;
			}
			set {
				this.enabled = value;
				if (EnabledChanged != null) {
					EnabledChanged(this, new EventArgs());
				}
			}
		}

		/// <summary>
		/// Gets or sets the toolbar button image.
		/// </summary>
		public Image Image
		{
			get {
				return this.image;
			}
			set {
				this.image = value;
			}
		}

		/// <summary>
		/// Gets or sets the text of the toolbar button's tooltip.
		/// </summary>
		public string ToolTip
		{
			get {
				return this.toolTip;
			}
			set {
				this.toolTip = value;
			}
		}

		/// <summary>
		/// Gets or sets a value that indicates whether the toolbar button acts
		/// as a seperator.
		/// </summary>
		public bool IsSeparator
		{
			get {
				return this.separator;
			}
			set {
				this.separator = value;
			}
		}

		/// <summary>
		/// Creates a toolbar separator.
		/// </summary>
		public static SimpleToolBarButton Separator
		{
			get {
				return new SimpleToolBarButton();
			}
		}
	}
}
